import bpy
from math import radians

def add_shrinkwrap(obj, target, vgroup):
    shrinkwrap = obj.modifiers.new(name="Shrinkwrap", type="SHRINKWRAP")

    shrinkwrap.target = target
    shrinkwrap.wrap_method = 'NEAREST_VERTEX'
    shrinkwrap.vertex_group = vgroup
    shrinkwrap.show_expanded = False
    shrinkwrap.show_on_cage = True

def add_boolean(obj, operator, method='DIFFERENCE', solver='FAST'):
    boolean = obj.modifiers.new(name=method.title(), type="BOOLEAN")

    boolean.object = operator
    boolean.operation = 'DIFFERENCE' if method == 'SPLIT' else method
    boolean.show_in_editmode = True
    boolean.show_expanded = False

    if method == 'SPLIT':
        boolean.show_viewport = False

    boolean.solver = solver

    return boolean

def add_displace(obj, name="Displace", mid_level=0, strength=0):
    displace = obj.modifiers.new(name=name, type="DISPLACE")
    displace.mid_level = mid_level
    displace.strength = strength

    displace.show_expanded = False

    return displace

def add_auto_smooth(obj, angle=20):
    with bpy.context.temp_override(object=obj):
        bpy.ops.object.modifier_add_node_group(asset_library_type='ESSENTIALS', asset_library_identifier="", relative_asset_identifier="geometry_nodes/smooth_by_angle.blend/NodeTree/Smooth by Angle")

        mod = get_auto_smooth(obj)

        if mod:
            mod.show_expanded = False
            mod.name = "Auto Smooth"

            mod['Input_1'] = radians(angle)
            mod.node_group.interface_update(bpy.context)
            return mod

        elif (ng := bpy.data.node_groups.get('Smooth by Angle')) and (mods := [mod for mod in obj.modifiers if mod.type == 'NODES' and not mod.node_group]):
            print("WARNING: Blender says: 'Warning: Asset loading is unfinished' (probably)")
            print("         But empty geo node mod is present, and 'Smooth by Angle' node group is present in the file already too")

            mod = mods[0]
            mod.node_group = ng

            mod.show_expanded = False
            mod.name = "Auto Smooth"

            mod['Input_1'] = radians(angle)
            mod.node_group.interface_update(bpy.context)
            return mod

        else:
            print("WARNING: Blender says: 'Warning: Asset loading is unfinished' (probably)")
            ng = bpy.data.node_groups.get('Smooth by Angle')
            mods = [mod for mod in obj.modifiers if mod.type == 'NODES' and not mod.node_group]

            print("         node group:", ng)
            print("empty geo node mods:", mods)
            print("                     TODO!")

def get_auto_smooth(obj):
    if (mod := obj.modifiers.get('Auto Smooth', None)) and mod.type == 'NODES':
        return mod

    elif (mod := obj.modifiers.get('Smooth by Angle', None)) and mod.type == 'NODES':
        return mod
    
    else:
        mods = [mod for mod in obj.modifiers if mod.type == 'NODES' and (ng := mod.node_group) and ng.name.startswith('Smooth by Angle')]

        if mods:
            return mods[0]

def apply_mod(modname):
    bpy.ops.object.modifier_apply(modifier=modname)

def remove_mod(mod):
    obj = mod.id_data
    obj.modifiers.remove(mod)

def get_mod_obj(mod):
    if mod.type in ['BOOLEAN', 'HOOK', 'LATTICE', 'DATA_TRANSFER']:
        return mod.object
    elif mod.type == 'MIRROR':
        return mod.mirror_object
    elif mod.type == 'ARRAY':
        return mod.offset_object

def get_mod_objects(obj, mod_objects, mod_types=['BOOLEAN', 'MIRROR', 'ARRAY', 'HOOK', 'LATTICE', 'DATA_TRANSFER'], recursive=True, depth=0, debug=False):
    depthstr = " " * depth

    if debug:
        print(f"\n{depthstr}{obj.name}")

    for mod in obj.modifiers:
        if mod.type in mod_types:
            mod_obj = get_mod_obj(mod)

            print(f" {depthstr}mod: {mod.name} | obj: {mod_obj.name if mod_obj else mod_obj}")

            if mod_obj:
                if mod_obj not in mod_objects:
                    mod_objects.append(mod_obj)

                if recursive:
                    get_mod_objects(mod_obj, mod_objects, mod_types=mod_types, recursive=True, depth=depth + 1, debug=debug)

def move_mod(mod, index=0):
    obj = mod.id_data
    current_index = list(obj.modifiers).index(mod)

    if current_index != index:
        obj.modifiers.move(current_index, index)

def sort_mod(mod):
    def is_boolean(mod):
        return mod.type == 'BOOLEAN'

    def is_mirror(mod):
        return mod.type == 'MIRROR'

    def is_array(mod):
        return mod.type == 'ARRAY'

    def is_auto_smooth(mod):
        return mod.type == 'NODES' and mod.node_group and 'Smooth by Angle' in mod.node_group.name

    def should_preceed(mod, prev_mod):
        if is_boolean(mod):
            return any([is_mirror(prev_mod), is_array(prev_mod), is_auto_smooth(prev_mod)])

        elif is_auto_smooth(mod):
            return any([is_mirror(prev_mod), is_array(prev_mod)])

    if any([is_boolean(mod), is_auto_smooth(mod)]):
        obj = mod.id_data
        mods = list(obj.modifiers)

        if len(mods) > 1:

            move_mod(mod, len(mods) - 1)

            index = len(mods) - 1

            while index:
                index -= 1
                prev_mod = obj.modifiers[index]

                if should_preceed(mod, prev_mod):

                    if index == 0:
                        move_mod(mod, index)
                    continue

                else:
                    move_mod(mod, index + 1)
                    break
